@@ -3,7 +3,7 @@ import struct
 import sys
 import traceback
 from enum import Enum
-from typing import List, Union, NamedTuple
+from typing import List, Union, NamedTuple, Callable
 import inspect
 from collections import OrderedDict
 from dataclasses import dataclass, field
@@ -39,22 +39,41 @@ class ExceptionType(Enum):
     ContextSwitch = 3
     Terminate = 4
 
+class MemoryViolation(Enum):
+    Unknown = 0
+    ReadUnmapped = 1
+    WriteUnmapped = 2
+    ExecuteUnmapped = 3
+    ReadProtect = 4
+    WriteProtect = 5
+    ExecuteProtect = 6
+    ReadUnaligned = 7
+    WriteUnaligned = 8
+    ExecuteUnaligned = 9
+
 @dataclass
 class ExceptionInfo:
     type: ExceptionType = ExceptionType.NoException
-    memory_access: int = 0  # refers to UC_MEM_* values
+    # type == ExceptionType.Memory
+    memory_violation: MemoryViolation = MemoryViolation.Unknown
     memory_address: int = 0
     memory_size: int = 0
     memory_value: int = 0
+    # type == ExceptionType.Interrupt
     interrupt_number: int = 0
+
+    # Internal state
+    _handling: bool = False
+
+@dataclass
+class UnicornExceptionInfo(ExceptionInfo):
+    final: bool = False
     code_hook_h: Optional[int] = None  # represents a `unicorn.uc_hook_h` value (from uc.hook_add)
     context: Optional[unicorn.UcContext] = None
     tb_start: int = 0
     tb_size: int = 0
     tb_icount: int = 0
     step_count: int = 0
-    final: bool = False
-    handling: bool = False
 
     def __str__(self):
         return f"{self.type}, ({hex(self.tb_start)}, {hex(self.tb_size)}, {self.tb_icount})"
@@ -296,6 +315,7 @@ class Dumpulator(Architecture):
         self.debug(f"total commit: {hex(self._pages.total_commit)}, pages: {self._pages.total_commit // PAGE_SIZE}")
         self._setup_modules()
         self.syscalls = []
+        self.win32k_syscalls = []
         self._setup_syscalls()
         self._setup_emulator(thread)
         self.handles = HandleManager()
@@ -304,8 +324,9 @@ class Dumpulator(Architecture):
         self.kill_me = None
         self.exit_code = None
         self.exports = self._all_exports()
-        self.exception = ExceptionInfo()
-        self.last_exception: Optional[ExceptionInfo] = None
+        self._exception = UnicornExceptionInfo()
+        self._last_exception: Optional[UnicornExceptionInfo] = None
+        self._exception_hook: Optional[Callable[[ExceptionInfo], Optional[int]]] = None
         if not self._quiet:
             print("Memory map:")
             self.print_memory()
@@ -340,7 +361,7 @@ class Dumpulator(Architecture):
             thread = self._minidump.threads.threads[i]
             if thread.ThreadId == thread_id:
                 return thread
-        raise Exception(f"Thread 0x{thread_id:x} ({thread_id}) not found!")
+        raise Exception(f"Thread {hex(thread_id)} ({thread_id}) not found!")
 
     def debug(self, message: str):
         if self._debug:
@@ -419,7 +440,7 @@ class Dumpulator(Architecture):
         seg: minidump.MinidumpMemorySegment
         for seg in self._minidump.memory_segments_64.memory_segments:
             emu_addr = seg.start_virtual_address & self.addr_mask
-            self.debug(f"initialize base: 0x{emu_addr:x}, size: 0x{seg.size:x}")
+            self.debug(f"initialize base: {hex(emu_addr)}, size: {hex(seg.size)}")
             memory.move(seg.start_virtual_address)
             assert memory.current_position == seg.start_virtual_address
             data = memory.read(seg.size)
@@ -508,11 +529,11 @@ class Dumpulator(Architecture):
         if self.wow64:
             self.memory.set_region_info(self.peb - PAGE_SIZE, "WoW64 PEB", size=PAGE_SIZE)
 
-        self.info(f"TEB: 0x{self.teb:x}, PEB: 0x{self.peb:x}")
-        self.info(f"  ConsoleHandle: 0x{self.console_handle:x}")
-        self.info(f"  StandardInput: 0x{self.stdin_handle:x}")
-        self.info(f"  StandardOutput: 0x{self.stdout_handle:x}")
-        self.info(f"  StandardError: 0x{self.stderr_handle:x}")
+        self.info(f"TEB: {hex(self.teb)}, PEB: {hex(self.peb)}")
+        self.info(f"  ConsoleHandle: {hex(self.console_handle)}")
+        self.info(f"  StandardInput: {hex(self.stdin_handle)}")
+        self.info(f"  StandardOutput: {hex(self.stdout_handle)}")
+        self.info(f"  StandardError: {hex(self.stderr_handle)}")
 
         process_heaps = []
         for i in range(0, min(number_of_heaps, 0x1000)):
@@ -779,13 +800,13 @@ class Dumpulator(Architecture):
     def _setup_syscalls(self):
         # Load the ntdll module from memory
         ntdll = self.modules["ntdll.dll"]
-        syscalls = []
+        nt_syscalls = []
         for export in ntdll.exports:
             if export.name and export.name.startswith("Zw"):
-                syscalls.append((export.address, export.name))
+                nt_syscalls.append((export.address, export.name))
             elif export.name == "Wow64Transition":
                 patch_addr = self.read_ptr(export.address)
-                self.info(f"Patching Wow64Transition: {export.address:x} -> {patch_addr:x}")
+                self.info(f"Patching Wow64Transition: {hex(export.address)} -> {hex(patch_addr)}")
                 # See: https://opcode0x90.wordpress.com/2007/05/18/kifastsystemcall-hook/
                 # mov edx, esp; sysenter; ret
                 KiFastSystemCall = b"\x8B\xD4\x0F\x34\x90\x90\xC3"
@@ -795,20 +816,46 @@ class Dumpulator(Architecture):
             elif export.name == "LdrLoadDll":
                 self.LdrLoadDll = export.address
 
-        syscalls.sort()
-        for index, (rva, name) in enumerate(syscalls):
-            cb = syscall_functions.get(name, None)
-            argcount = 0
-            if cb:
-                argspec = inspect.getfullargspec(cb)
-                argcount = len(argspec.args) - 1
-            self.syscalls.append((name, cb, argcount))
+        def add_syscalls(syscalls, table):
+            # The index when sorting by RVA is the syscall index
+            syscalls.sort()
+            for index, (rva, name) in enumerate(syscalls):
+                cb = syscall_functions.get(name, None)
+                argcount = 0
+                if cb:
+                    argspec = inspect.getfullargspec(cb)
+                    argcount = len(argspec.args) - 1
+                table.append((name, cb, argcount))
+
+        add_syscalls(nt_syscalls, self.syscalls)
+
+        # Get the syscalls for win32u
+        win32u = self.modules.find("win32u.dll")
+        if win32u is not None:
+            win32k_syscalls = []
+            for export in win32u.exports:
+                if export.name and export.name.startswith("Nt"):
+                    win32k_syscalls.append((export.address, export.name))
+
+            add_syscalls(win32k_syscalls, self.win32k_syscalls)
+
 
     def push(self, value):
         csp = self.regs.csp - self.ptr_size()
         self.write_ptr(csp, value)
         self.regs.csp = csp
 
+    def pop(self):
+        csp = self.regs.csp
+        value = self.read_ptr(csp)
+        self.regs.csp = csp + self.ptr_size()
+        return value
+
+    def ret(self, imm=0):
+        return_address = self.pop()
+        self.regs.csp -= imm
+        return return_address
+
     def read(self, addr, size):
         if not isinstance(addr, int):
             addr = int(addr)
@@ -869,15 +916,28 @@ class Dumpulator(Architecture):
         self.memory.commit(self.memory.align_page(ptr), self.memory.align_page(size))
         return ptr
 
-    def handle_exception(self):
-        assert not self.exception.handling
-        self.exception.handling = True
+    def set_exception_hook(self, exception_hook: Optional[Callable[[ExceptionInfo], Optional[int]]]):
+        previous_hook = self._exception_hook
+        self._exception_hook = exception_hook
+        return previous_hook
 
-        if self.exception.type == ExceptionType.ContextSwitch:
+    def handle_exception(self):
+        assert not self._exception._handling
+        self._exception._handling = True
+
+        if self._exception_hook is not None:
+            hook_result = self._exception_hook(self._exception)
+            if hook_result is not None:
+                # Clear the pending exception
+                self._last_exception = self._exception
+                self._exception = UnicornExceptionInfo()
+                return hook_result
+
+        if self._exception.type == ExceptionType.ContextSwitch:
             self.info(f"context switch, cip: {hex(self.regs.cip)}")
             # Clear the pending exception
-            self.last_exception = self.exception
-            self.exception = ExceptionInfo()
+            self._last_exception = self._exception
+            self._exception = UnicornExceptionInfo()
             # NOTE: the context has already been restored using context_restore in the caller
             return self.regs.cip
 
@@ -921,7 +981,7 @@ rsp in KiUserExceptionDispatcher:
 
         csp = self.regs.csp - allocation_size
         self.write(csp, allocation_size * b"\x69")  # fill stuff with 0x69 for debugging
-        self.info(f"old csp: {self.regs.csp:x}, new csp: {csp:x}")
+        self.info(f"old csp: {hex(self.regs.csp)}, new csp: {hex(csp)}")
         context_size = ctypes.sizeof(context_type)
         context = context_type.from_buffer(self.read(csp, context_size))
         context.ContextFlags = context_flags
@@ -934,22 +994,23 @@ rsp in KiUserExceptionDispatcher:
         context_ex.XState.Offset = 0xF0 if self._x64 else 0x20
         context_ex.XState.Length = 0x160 if self._x64 else 0x140
         record = record_type()
-        if self.exception.type == ExceptionType.Memory:
-            record.ExceptionCode = 0xC0000005
+        alignment_violations = [MemoryViolation.ReadUnaligned, MemoryViolation.WriteUnaligned, MemoryViolation.ExecuteUnaligned]
+        if self._exception.type == ExceptionType.Memory and self._exception.memory_violation not in alignment_violations:
+            record.ExceptionCode = STATUS_ACCESS_VIOLATION
             record.ExceptionFlags = 0
             record.ExceptionAddress = self.regs.cip
             record.NumberParameters = 2
             types = {
-                UC_MEM_READ_UNMAPPED: EXCEPTION_READ_FAULT,
-                UC_MEM_WRITE_UNMAPPED: EXCEPTION_WRITE_FAULT,
-                UC_MEM_FETCH_UNMAPPED: EXCEPTION_READ_FAULT,
-                UC_MEM_READ_PROT: EXCEPTION_READ_FAULT,
-                UC_MEM_WRITE_PROT: EXCEPTION_WRITE_FAULT,
-                UC_MEM_FETCH_PROT: EXCEPTION_EXECUTE_FAULT,
+                MemoryViolation.ReadUnmapped: EXCEPTION_READ_FAULT,
+                MemoryViolation.WriteUnmapped: EXCEPTION_WRITE_FAULT,
+                MemoryViolation.ExecuteUnmapped: EXCEPTION_READ_FAULT,
+                MemoryViolation.ReadProtect: EXCEPTION_READ_FAULT,
+                MemoryViolation.WriteProtect: EXCEPTION_WRITE_FAULT,
+                MemoryViolation.ExecuteProtect: EXCEPTION_EXECUTE_FAULT,
             }
-            record.ExceptionInformation[0] = types[self.exception.memory_access]
-            record.ExceptionInformation[1] = self.exception.memory_address
-        elif self.exception.type == ExceptionType.Interrupt and self.exception.interrupt_number == 3:
+            record.ExceptionInformation[0] = types[self._exception.memory_violation]
+            record.ExceptionInformation[1] = self._exception.memory_address
+        elif self._exception.type == ExceptionType.Interrupt and self._exception.interrupt_number == 3:
             if self._x64:
                 context.Rip -= 1  # TODO: long int3 and prefixes
                 record.ExceptionCode = 0x80000003
@@ -963,11 +1024,11 @@ rsp in KiUserExceptionDispatcher:
                 record.ExceptionAddress = context.Eip
                 record.NumberParameters = 1
         else:
-            raise NotImplementedError(f"{self.exception}")  # TODO: implement
+            raise NotImplementedError(f"{self._exception}")  # TODO: implement
 
         # Clear the pending exception
-        self.last_exception = self.exception
-        self.exception = ExceptionInfo()
+        self._last_exception = self._exception
+        self._exception = UnicornExceptionInfo()
 
         def write_stack(cur_ptr: int, data: bytes):
             self.write(cur_ptr, data)
@@ -997,26 +1058,26 @@ rsp in KiUserExceptionDispatcher:
 
     def start(self, begin, end=0xffffffffffffffff, count=0) -> None:
         # Clear exceptions before starting
-        self.exception = ExceptionInfo()
+        self._exception = UnicornExceptionInfo()
         emu_begin = begin
         emu_until = end
         emu_count = count
         while True:
             try:
-                if self.exception.type != ExceptionType.NoException:
-                    if self.exception.final:
+                if self._exception.type != ExceptionType.NoException:
+                    if self._exception.final:
                         # Restore the context (unicorn might mess with it before stopping)
-                        if self.exception.context is not None:
-                            self._uc.context_restore(self.exception.context)
+                        if self._exception.context is not None:
+                            self._uc.context_restore(self._exception.context)
 
-                        if self.exception.type == ExceptionType.Terminate:
+                        if self._exception.type == ExceptionType.Terminate:
                             if self.exit_code is not None:
-                                self.info(f"exit code: {self.exit_code:x}")
+                                self.info(f"exit code: {hex(self.exit_code)}")
                             break
 
                         try:
                             emu_begin = self.handle_exception()
-                        except:
+                        except Exception:
                             traceback.print_exc()
                             self.error(f"exception during exception handling (stack overflow?)")
                             break
@@ -1024,36 +1085,36 @@ rsp in KiUserExceptionDispatcher:
                         emu_count = 0
                     else:
                         # If this happens there was an error restarting simulation
-                        assert self.exception.step_count == 0
+                        assert self._exception.step_count == 0
 
                         # Hook should be installed at this point
-                        assert self.exception.code_hook_h is not None
+                        assert self._exception.code_hook_h is not None
 
                         # Restore the context (unicorn might mess with it before stopping)
-                        assert self.exception.context is not None
-                        self._uc.context_restore(self.exception.context)
+                        assert self._exception.context is not None
+                        self._uc.context_restore(self._exception.context)
 
                         # Restart emulation
                         self.info(f"restarting emulation to handle exception...")
                         emu_begin = self.regs.cip
                         emu_until = 0xffffffffffffffff
-                        emu_count = self.exception.tb_icount + 1
+                        emu_count = self._exception.tb_icount + 1
 
-                self.info(f"emu_start({emu_begin:x}, {emu_until:x}, {emu_count})")
+                self.info(f"emu_start({hex(emu_begin)}, {hex(emu_until)}, {emu_count})")
                 self.kill_me = None
                 self._uc.emu_start(emu_begin, until=emu_until, count=emu_count)
-                self.info(f'emulation finished, cip = {self.regs.cip:x}')
+                self.info(f'emulation finished, cip = {hex(self.regs.cip)}')
                 if self.exit_code is not None:
-                    self.info(f"exit code: {self.exit_code:x}")
+                    self.info(f"exit code: {hex(self.exit_code)}")
                 break
             except UcError as err:
                 if self.kill_me is not None and type(self.kill_me) is not UcError:
                     raise self.kill_me
-                if self.exception.type != ExceptionType.NoException:
+                if self._exception.type != ExceptionType.NoException:
                     # Handle the exception outside of the except handler
                     continue
                 else:
-                    self.error(f'error: {err}, cip = {self.regs.cip:x}')
+                    self.error(f'error: {err}, cip = {hex(self.regs.cip)}')
                     traceback.print_exc()
                 break
 
@@ -1062,7 +1123,7 @@ rsp in KiUserExceptionDispatcher:
             self.exit_code = None
             if exit_code is not None:
                 self.exit_code = int(exit_code)
-        except:
+        except Exception:
             traceback.print_exc()
             self.error("Invalid type passed to exit_code!")
         self._uc.emu_stop()
@@ -1204,8 +1265,8 @@ rsp in KiUserExceptionDispatcher:
 
 def _hook_code_exception(uc: Uc, address, size, dp: Dumpulator):
     try:
-        dp.info(f"exception step: {address:x}[{size}]")
-        ex = dp.exception
+        dp.info(f"exception step: {hex(address)}[{size}]")
+        ex = dp._exception
         ex.step_count += 1
         if ex.step_count >= ex.tb_icount:
             raise Exception("Stepped past the basic block without reaching exception")
@@ -1219,18 +1280,27 @@ def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
         return True
 
     fetch_accesses = [UC_MEM_FETCH, UC_MEM_FETCH_PROT, UC_MEM_FETCH_UNMAPPED]
-    if access == UC_MEM_FETCH_UNMAPPED and address >= FORCE_KILL_ADDR - 0x10 and address <= FORCE_KILL_ADDR + 0x10 and dp.kill_me is not None:
-        dp.error(f"forced exit memory operation {access} of {address:x}[{size:x}] = {value:X}")
+    if access == UC_MEM_FETCH_UNMAPPED and FORCE_KILL_ADDR - 0x10 <= address <= FORCE_KILL_ADDR + 0x10 and dp.kill_me is not None:
+        dp.error(f"forced exit memory operation {access} of {hex(address)}[{hex(size)}] = {hex(value)}")
         return False
-    if dp.exception.final and access in fetch_accesses:
+    if dp._exception.final and access in fetch_accesses:
         dp.info(f"fetch from {hex(address)}[{size}] already reported")
         return False
     # TODO: figure out why when you start executing at 0 this callback is triggered more than once
     try:
+        violation = {
+            UC_MEM_READ_UNMAPPED: MemoryViolation.ReadUnmapped,
+            UC_MEM_WRITE_UNMAPPED: MemoryViolation.WriteUnmapped,
+            UC_MEM_FETCH_UNMAPPED: MemoryViolation.ExecuteUnmapped,
+            UC_MEM_READ_PROT: MemoryViolation.ReadProtect,
+            UC_MEM_WRITE_PROT: MemoryViolation.WriteProtect,
+            UC_MEM_FETCH_PROT: MemoryViolation.ExecuteProtect,
+        }.get(access, MemoryViolation.Unknown)
+        assert violation != MemoryViolation.Unknown, f"Unexpected memory access {access}"
         # Extract exception information
-        exception = ExceptionInfo()
+        exception = UnicornExceptionInfo()
         exception.type = ExceptionType.Memory
-        exception.memory_access = access
+        exception.memory_violation = violation
         exception.memory_address = address
         exception.memory_size = size
         exception.memory_value = value
@@ -1242,14 +1312,14 @@ def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
             exception.tb_icount = tb.icount
 
         # Print exception info
-        final = dp.trace or dp.exception.code_hook_h is not None
+        final = dp.trace or dp._exception.code_hook_h is not None
         info = "final" if final else "initial"
         if access == UC_MEM_READ_UNMAPPED:
-            dp.error(f"{info} unmapped read from {address:x}[{size:x}], cip = {dp.regs.cip:x}, exception: {exception}")
+            dp.error(f"{info} unmapped read from {hex(address)}[{hex(size)}], cip = {hex(dp.regs.cip)}, exception: {exception}")
         elif access == UC_MEM_WRITE_UNMAPPED:
-            dp.error(f"{info} unmapped write to {address:x}[{size:x}] = {value:x}, cip = {dp.regs.cip:x}")
+            dp.error(f"{info} unmapped write to {hex(address)}[{hex(size)}] = {hex(value)}, cip = {hex(dp.regs.cip)}")
         elif access == UC_MEM_FETCH_UNMAPPED:
-            dp.error(f"{info} unmapped fetch of {address:x}[{size:x}], cip = {dp.regs.rip:x}, cs = {dp.regs.cs:x}")
+            dp.error(f"{info} unmapped fetch of {hex(address)}[{hex(size)}], cip = {hex(dp.regs.rip)}, cs = {hex(dp.regs.cs)}")
         else:
             names = {
                 UC_MEM_READ: "UC_MEM_READ", # Memory is read from
@@ -1263,30 +1333,30 @@ def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
                 UC_MEM_FETCH_PROT: "UC_MEM_FETCH_PROT", # Fetch from non-executable, but mapped, memory
                 UC_MEM_READ_AFTER: "UC_MEM_READ_AFTER", # Memory is read from (successful access)
             }
-            dp.error(f"{info} unsupported access {names.get(access, str(access))} of {address:x}[{size:x}] = {value:X}, cip = {dp.regs.cip:x}")
+            dp.error(f"{info} unsupported access {names.get(access, str(access))} of {hex(address)}[{hex(size)}] = {hex(value)}, cip = {hex(dp.regs.cip)}")
 
         if final:
             # Make sure this is the same exception we expect
             if not dp.trace:
-                assert access == dp.exception.memory_access
-                assert address == dp.exception.memory_address
-                assert size == dp.exception.memory_size
-                assert value == dp.exception.memory_value
+                assert violation == dp._exception.memory_violation
+                assert address == dp._exception.memory_address
+                assert size == dp._exception.memory_size
+                assert value == dp._exception.memory_value
 
                 # Delete the code hook
-                uc.hook_del(dp.exception.code_hook_h)
-                dp.exception.code_hook_h = None
+                uc.hook_del(dp._exception.code_hook_h)
+                dp._exception.code_hook_h = None
 
             # At this point we know for sure the context is correct so we can report the exception
-            dp.exception = exception
-            dp.exception.final = True
+            dp._exception = exception
+            dp._exception.final = True
 
             # Stop emulation (we resume it on KiUserExceptionDispatcher later)
             dp.stop()
             return False
 
         # There should not be an exception active
-        assert dp.exception.type == ExceptionType.NoException
+        assert dp._exception.type == ExceptionType.NoException
 
         # Remove the translation block cache for this block
         # Without doing this single stepping the block won't work
@@ -1298,7 +1368,7 @@ def _hook_mem(uc: Uc, access, address, size, value, dp: Dumpulator):
         exception.code_hook_h = uc.hook_add(UC_HOOK_CODE, _hook_code_exception, user_data=dp)
 
         # Store the exception info
-        dp.exception = exception
+        dp._exception = exception
 
         # Stop emulation (we resume execution later)
         dp.stop()
@@ -1360,14 +1430,14 @@ def _hook_code(uc: Uc, address, size, dp: Dumpulator):
     elif module:
         address_name = " " + module
 
-    line = f"0x{address:x}{address_name}|"
+    line = f"{hex(address)}{address_name}|"
     if instr is not None:
         line += instr.mnemonic
         if instr.op_str:
             line += " "
             line += instr.op_str
         for reg in _get_regs(instr):
-            line += f"|{reg}=0x{dp.regs.__getattr__(reg):x}"
+            line += f"|{reg}={hex(dp.regs.__getattr__(reg))}"
         if instr.mnemonic in {"syscall", "sysenter"}:
             line += f"|sequence_id=[{dp.sequence_id}]"
     else:
@@ -1425,7 +1495,7 @@ def _arg_type_string(arg):
 def _hook_interrupt(uc: Uc, number, dp: Dumpulator):
     try:
         # Extract exception information
-        exception = ExceptionInfo()
+        exception = UnicornExceptionInfo()
         exception.type = ExceptionType.Interrupt
         exception.interrupt_number = number
         exception.context = uc.context_save()
@@ -1440,14 +1510,14 @@ def _hook_interrupt(uc: Uc, number, dp: Dumpulator):
             description = interrupt_names[number]
         else:
             description = f"IRQ {number - 32}"
-        dp.error(f"interrupt {number} ({description}), cip = {dp.regs.cip:x}, cs = {dp.regs.cs:x}")
+        dp.error(f"interrupt {number} ({description}), cip = {hex(dp.regs.cip)}, cs = {hex(dp.regs.cs)}")
 
         # There should not be an exception active
-        assert dp.exception.type == ExceptionType.NoException
+        assert dp._exception.type == ExceptionType.NoException
 
         # At this point we know for sure the context is correct so we can report the exception
-        dp.exception = exception
-        dp.exception.final = True
+        dp._exception = exception
+        dp._exception.final = True
     except AssertionError as err:
         traceback.print_exc()
         raise err
@@ -1464,19 +1534,36 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
     # Flush the trace for easier debugging
     if dp.trace is not None:
         dp.trace.flush()
-    index = dp.regs.cax & 0xffff
-    if index < len(dp.syscalls):
-        name, syscall_impl, argcount = dp.syscalls[index]
+
+    # Extract the table and function number from eax
+    service_number = dp.regs.cax & 0xffff
+    table_number = (service_number >> 12) & 0xf  # 0: ntoskrnl, 1: win32k
+    function_index = service_number & 0xfff
+    if table_number == 0:
+        table = dp.syscalls
+        table_prefix = ""
+    elif table_number == 1:
+        table = dp.win32k_syscalls
+        table_prefix = "win32k "
+    else:
+        table = []
+        table_prefix = f"unknown:{table_number} "
+
+    if function_index < len(table):
+        name, syscall_impl, argcount = table[function_index]
         if syscall_impl:
             argspec = inspect.getfullargspec(syscall_impl)
             args = []
 
             def syscall_arg(index):
+                # There is an extra call that adds a return address to the stack
+                if dp.wow64:
+                    index += 1
                 if index == 0 and dp.ptr_size() == 8:
                     return dp.regs.r10
                 return dp.args[index]
 
-            dp.info(f"[{dp.sequence_id}] syscall (index: {hex(index)}): {name}(")
+            dp.info(f"[{dp.sequence_id}] {table_prefix}syscall: {name}( /* index: {hex(service_number)} */")
             for i in range(0, argcount):
                 argname = argspec.args[1 + i]
                 argtype = argspec.annotations[argname]
@@ -1499,9 +1586,9 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                     argvalue = argtype(dp, argvalue)
                 elif issubclass(argtype, Enum):
                     try:
-                        argvalue = argtype(dp.args[i] & 0xFFFFFFFF)
+                        argvalue = argtype(argvalue & 0xFFFFFFFF)
                     except KeyError as x:
-                        raise Exception(f"Unknown enum value {dp.args[i]} for {type(argtype)}") from None
+                        raise Exception(f"Unknown enum value {argvalue} for {type(argtype)}") from None
                 else:
                     argvalue = argtype(argvalue)
                 args.append(argvalue)
@@ -1516,10 +1603,10 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
                 status = syscall_impl(dp, *args)
                 if isinstance(status, ExceptionInfo):
                     print("context switch, stopping emulation")
-                    dp.exception = status
+                    dp._exception = status
                     raise dp.raise_kill(UcError(UC_ERR_EXCEPTION)) from None
                 else:
-                    dp.info(f"status = {status:x}")
+                    dp.info(f"status = {hex(status)}")
                     dp.regs.cax = status
                     if dp.x64:
                         dp.regs.rcx = dp.regs.cip + 2
@@ -1535,9 +1622,9 @@ def _hook_syscall(uc: Uc, dp: Dumpulator):
             finally:
                 dp.sequence_id += 1
         else:
-            raise dp.raise_kill(NotImplementedError(f"syscall index: {index:x} -> {name} not implemented!")) from None
+            raise dp.raise_kill(NotImplementedError(f"{table_prefix}syscall {hex(service_number)} -> {name} not implemented!")) from None
     else:
-        raise dp.raise_kill(IndexError(f"syscall index {index:x} out of range")) from None
+        raise dp.raise_kill(IndexError(f"{table_prefix}syscall {hex(service_number)} (index: {hex(function_index)}) out of range")) from None
 
 def _emulate_unsupported_instruction(dp: Dumpulator, instr: CsInsn):
     if instr.id == X86_INS_RDRAND:
@@ -1560,17 +1647,17 @@ def _hook_invalid(uc: Uc, dp: Dumpulator):
     if dp.kill_me:
         dp.error(f"terminating emulation...")
         return False
-    dp.error(f"invalid instruction at {address:x}")
+    dp.error(f"invalid instruction at {hex(address)}")
     try:
         code = dp.read(address, 15)
         instr = next(dp.cs.disasm(code, address, 1))
         if _emulate_unsupported_instruction(dp, instr):
             # Resume execution with a context switch
-            assert dp.exception.type == ExceptionType.NoException
-            exception = ExceptionInfo()
+            assert dp._exception.type == ExceptionType.NoException
+            exception = UnicornExceptionInfo()
             exception.type = ExceptionType.ContextSwitch
             exception.final = True
-            dp.exception = exception
+            dp._exception = exception
             return False  # NOTE: returning True would stop emulation
     except StopIteration:
         pass  # Unsupported instruction
